#!/usr/bin/env bash
# nbdkit
# Copyright Red Hat
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Red Hat nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY RED HAT AND CONTRIBUTORS ''AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RED HAT OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

# Test the --log=/path option appends.

source ./functions.sh
set -e
set -x
set -u

export LANG=C

requires_run
requires_plugin null
requires_filter error
requires_nbdinfo

logfile="$PWD/log-append.out"
rm -f $logfile
cleanup_fn rm -f $logfile

# Write some pre-existing content to the log file, which should not be
# overwritten.
echo preexiting0  >$logfile
echo preexiting1 >>$logfile
echo preexiting2 >>$logfile

nbdkit --log="$logfile" null size=10k \
       --filter=error error=EPERM error-rate=1 \
       --run 'nbdinfo "$uri" ||:'

nbdkit --log="$logfile" null size=10k \
       --filter=error error=EIO error-rate=1 \
       --run 'nbdinfo "$uri" ||:'

# Check the log file exists.
cat $logfile

# Check the pre-existing content is still present.
grep preexiting0 $logfile
grep preexiting1 $logfile
grep preexiting2 $logfile

# Check both instances of nbdkit logged to the file.
grep EPERM $logfile
grep EIO $logfile
