/* nbdkit
 * Copyright Red Hat
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of Red Hat nor the names of its contributors may be
 * used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY RED HAT AND CONTRIBUTORS ''AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RED HAT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Replacement for realpath for platforms which lack this function. */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#ifndef HAVE_REALPATH

#include "realpath.h"

#ifdef WIN32

/* Replacement realpath for Win32.
 *
 * Note this is not thread-safe, but should be fine for all the
 * uses in the server and ordinary plugins.
 */

#include <windows.h>

char *
realpath (const char *path, char *out)
{
  TCHAR buf[MAX_PATH];
  int r;

  r = GetFullPathNameA (path, MAX_PATH, buf, NULL);
  if (r == 0) {
    errno = GetLastError ();
    return NULL;
  }

  out = malloc (r + 1);
  if (out == NULL)
    return NULL;
  memcpy (out, buf, r);
  out[r] = '\0';

  return out;
}

#else /* !WIN32 */
#error "no replacement realpath is available on this platform"
#endif

#endif /* !HAVE_REALPATH */
