/* libnbd
 * Copyright Red Hat
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of Red Hat nor the names of its contributors may be
 * used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY RED HAT AND CONTRIBUTORS ''AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RED HAT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef LIBNBD_BENCH_H
#define LIBNBD_BENCH_H

#include <sys/time.h>

#define MICROSECONDS 1000000

struct bench {
  struct timeval start, stop;
};

static inline void
bench_start (struct bench *b)
{
  gettimeofday (&b->start, NULL);
}

static inline void
bench_stop (struct bench *b)
{
  gettimeofday (&b->stop, NULL);
}

static inline double
bench_sec (struct bench *b)
{
  struct timeval dt;

  dt.tv_sec = b->stop.tv_sec - b->start.tv_sec;
  dt.tv_usec = b->stop.tv_usec - b->start.tv_usec;

  if (dt.tv_usec < 0) {
    dt.tv_sec -= 1;
    dt.tv_usec += MICROSECONDS;
  }

  return ((double)dt.tv_sec * MICROSECONDS + dt.tv_usec) / MICROSECONDS;
}

#endif /* LIBNBD_BENCH_H */
