;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")
((autoconf-mode . ((indent-tabs-mode . nil)))
 (c-mode . ((indent-tabs-mode . nil)
	    (c-basic-offset . 2))))
