/* DO NOT EDIT!  Generated from ../docs/synopsis.txt */

const char *synopsis =
    "nbdkit [-4|--ipv4-only] [-6|--ipv6-only]\n"
    "       [-D|--debug PLUGIN|FILTER|nbdkit.FLAG=N]\n"
    "       [--exit-with-parent] [-e|--exportname EXPORTNAME]\n"
    "       [--filter=FILTER ...] [-f|--foreground]\n"
    "       [-g|--group GROUP] [-i|--ipaddr IPADDR] [--keepalive]\n"
    "       [--log=default|stderr|syslog|null|/path]\n"
    "       [--mask-handshake=MASK] [--name=NAME]\n"
    "       [-n|--newstyle] [--no-mc|--no-meta-contexts]\n"
    "       [--no-sr|--no-structured-replies] [-o|--oldstyle]\n"
    "       [-P|--pidfile PIDFILE] [-p|--port PORT] [--print-uri]\n"
    "       [-r|--readonly] [--run 'COMMAND ARGS ...']\n"
    "       [--selinux-label=LABEL] [-s|--single]\n"
    "       [--swap] [-t|--threads THREADS]\n"
    "       [--timeout=TIMEOUT] [--tls=off|on|require]\n"
    "       [--tls-certificates=/path/to/certificates]\n"
    "       [--tls-priority=PRIORITY]\n"
    "       [--tls-psk=/path/to/pskfile] [--tls-verify-peer]\n"
    "       [-U|--unix SOCKET|-] [-u|--user USER]\n"
    "       [-v|--verbose] [--vsock]\n"
    "       PLUGIN\n"
    "       [[KEY=]VALUE [KEY=VALUE [...]]] | @PATH\n"
    "\n"
    "nbdkit --dump-config\n"
    "\n"
    "nbdkit PLUGIN --dump-plugin\n"
    "\n"
    "nbdkit --help\n"
    "\n"
    "nbdkit [-V|--version]\n"
    "\n"
    "nbdkit [--short-options|--long-options]\n"
    ;
